@@ -1,12 +1,15 @@
 
 # -----------------------------------------------------------------------------
 # The user interface of the _operations is build by magicgui
+from typing import Union
+
 from qtpy.QtWidgets import QTableWidget, QTableWidgetItem
 from magicgui import magicgui
 from napari.layers import Image, Labels, Layer
 import pyclesperanto_prototype as cle
 
 plus_minus_1k = {'min': -1000, 'max': 1000}
+plus_10k = {'min': 0, 'max': 10000}
 
 def _call_operation_ignoring_to_many_arguments(operation, arguments):
     """
@@ -16,9 +19,30 @@ def _call_operation_ignoring_to_many_arguments(operation, arguments):
 
     import inspect
     sig = inspect.signature(operation)
-    if len(sig.parameters) < len(arguments):
-        arguments = arguments[:len(sig.parameters)]
-    operation(*arguments)
+
+    count = 0
+    for parameter in sig.parameters:
+        if sig.parameters[parameter].annotation in [cle.Image, int, str, float, bool]:
+            count = count + 1
+
+    if count < len(arguments):
+        arguments = arguments[:count]
+    return operation(*arguments)
+
+def label_parameters(operation, parameters):
+    import inspect
+    sig = inspect.signature(operation)
+
+    count = 0
+    for parameter in sig.parameters:
+        if sig.parameters[parameter].annotation in [int, str, float, bool]:
+            if len(parameters) > count:
+                parameters[count].label = parameter
+                parameters[count].text = parameter
+                count = count + 1
+    for i in range(count, len(parameters)):
+        parameters[i].label = ""
+        parameters[i].text = ""
 
 @magicgui(
     auto_call=True,
@@ -35,6 +59,8 @@ def denoise(input1: Image, operation_name: str = cle.gaussian_blur.__name__, x:
         cle_input = cle.push(input1.data)
         output = cle.create_like(cle_input)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [denoise.x, denoise.y, denoise.z])
         _call_operation_ignoring_to_many_arguments(operation, [cle_input, output, x, y, z])
         max_intensity = cle.maximum_of_all_pixels(output)
         if max_intensity == 0:
@@ -65,6 +91,8 @@ def background_removal(input1: Image, operation_name: str = cle.top_hat_box.__na
         cle_input = cle.push(input1.data)
         output = cle.create_like(cle_input)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [background_removal.x, background_removal.y, background_removal.z])
         _call_operation_ignoring_to_many_arguments(operation, [cle_input, output, x, y, z])
         max_intensity = cle.maximum_of_all_pixels(output)
         if max_intensity == 0:
@@ -95,6 +123,8 @@ def filter(input1: Image, operation_name: str = cle.gamma_correction.__name__, x
         cle_input = cle.push(input1.data)
         output = cle.create_like(cle_input)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [filter.x, filter.y, filter.z])
         _call_operation_ignoring_to_many_arguments(operation, [cle_input, output, x, y, z])
         max_intensity = cle.maximum_of_all_pixels(output)
         if max_intensity == 0:
@@ -126,7 +156,9 @@ def binarize(input1: Layer, operation_name : str = cle.threshold_otsu.__name__,
         cle_input1 = cle.push(input1.data)
         output = cle.create_like(cle_input1)
         operation = cle.operation(operation_name)
-        _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output, radius_x, radius_y, radius_y])
+        # update GUI
+        label_parameters(operation, [binarize.radius_x, binarize.radius_y, binarize.radius_z])
+        _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output, radius_x, radius_y, radius_z])
         output = cle.pull(output).astype(int)
 
         # show result in napari
@@ -176,15 +208,19 @@ def combine(input1: Layer, input2: Layer = None, operation_name: str = cle.binar
     auto_call=True,
     layout='vertical',
     input1={'label':'Image'},
-    operation_name={'label': 'Operation', 'choices':cle.operations(must_have_categories=['label', 'in assistant']).keys()}
+    operation_name={'label': 'Operation', 'choices':cle.operations(must_have_categories=['label', 'in assistant']).keys()},
+    a=plus_minus_1k,
+    b=plus_minus_1k
 )
-def label(input1: Layer, operation_name: str = cle.connected_components_labeling_box.__name__):
+def label(input1: Layer, operation_name: str = cle.connected_components_labeling_box.__name__, a : float = 2, b : float = 2):
     if input1 is not None:
         # execute operation
         cle_input1 = cle.push(input1.data)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [label.a, label.b])
         output = cle.create_like(cle_input1)
-        _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output])
+        _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output, a, b])
         output = cle.pull(output).astype(int)
 
         # show result in napari
@@ -210,6 +246,8 @@ def label_processing(input1: Labels, operation_name: str = cle.exclude_labels_on
         cle_input1 = cle.push(input1.data)
         output = cle.create_like(cle_input1)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [label_processing.min, label_processing.max])
         _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output, min, max])
         output = cle.pull(output).astype(int)
 
@@ -239,6 +277,8 @@ def label_measurements(input1: Image, input2: Labels = None, operation_name: str
         cle_input2 = cle.push(input2.data)
         output = cle.create_like(cle_input1)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [label_measurements.n])
         _call_operation_ignoring_to_many_arguments(operation, [cle_input1, cle_input2, output, n])
         max_intensity = cle.maximum_of_all_pixels(output)
         if max_intensity == 0:
@@ -269,6 +309,8 @@ def mesh(input1: Labels, operation_name : str = cle.draw_mesh_between_touching_l
         cle_input1 = cle.push(input1.data)
         output = cle.create_like(cle_input1)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [mesh.n])
         _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output, n])
         min_intensity = cle.minimum_of_all_pixels(output)
         max_intensity = cle.maximum_of_all_pixels(output)
@@ -299,6 +341,8 @@ def map(input1: Labels, operation_name: str = cle.label_pixel_count_map.__name__
         cle_input1 = cle.push(input1.data)
         output = cle.create_like(cle_input1)
         operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [map.n])
         _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output, n])
         max_intensity = cle.maximum_of_all_pixels(output)
         if max_intensity == 0:
@@ -336,3 +380,72 @@ def table_to_widget(table : dict) -> QTableWidget:
         for j, value in enumerate(table.get(column)):
             view.setItem(j + 1, i,  QTableWidgetItem(str(value)))
     return view
+
+# -----------------------------------------------------------------------------
+@magicgui(
+    auto_call=True,
+    layout='vertical',
+    input1={'label':'Image'},
+    operation_name={'label': 'Operation', 'choices':cle.operations(must_have_categories=['transform', 'in assistant']).keys()},
+    a=plus_minus_1k,
+    b=plus_minus_1k,
+    c=plus_minus_1k
+)
+def transform(input1: Layer, operation_name : str = cle.sub_stack.__name__, a : float = 0, b : float = 0, c : float = 0, d : bool = False, e : bool = False):
+    if input1 is not None:
+        # determine shift; todo: to this in a generic way
+        translate = input1.translate
+        if operation_name == cle.sub_stack.__name__:
+            translate = [input1.translate[0] + a, input1.translate[1], input1.translate[2]] # a corresponds to start_z
+        elif operation_name == cle.reduce_stack.__name__:
+            translate = [input1.translate[0] + b, input1.translate[1], input1.translate[2]] # b corresponds to offset
+
+        # execute operation
+        cle_input1 = cle.push(input1.data)
+        output = None
+        operation = cle.operation(operation_name)
+        # update GUI
+        label_parameters(operation, [transform.a, transform.b, transform.c, transform.d, transform.e])
+        output = _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output, a, b, c, d, e])
+        output = cle.pull(output).astype(int)
+
+        # show result in napari
+        if not hasattr(transform.self, 'layer'):
+            if isinstance(input1, Labels):
+                transform.self.viewer.add_labels(output, translate=translate)
+            else:
+                transform.self.viewer.add_image(output, translate=translate)
+        else:
+            transform.self.layer.data = output
+            transform.self.layer.contrast_limits = input1.contrast_limits
+            transform.self.layer.name = "Result of " + operation.__name__
+            transform.self.layer.translate = translate
+
+# -----------------------------------------------------------------------------
+@magicgui(
+    auto_call=True,
+    layout='vertical',
+    input1={'label':'Image'},
+    operation_name={'label': 'Operation', 'choices':cle.operations(must_have_categories=['projection', 'in assistant']).keys()}
+)
+def projection(input1: Layer, operation_name : str = cle.maximum_z_projection.__name__):
+    if input1 is not None:
+        # execute operation
+        cle_input1 = cle.push(input1.data)
+        output = None
+        operation = cle.operation(operation_name)
+        output = _call_operation_ignoring_to_many_arguments(operation, [cle_input1, output])
+        output = cle.pull(output).astype(int)
+
+
+        # show result in napari
+        if not hasattr(projection.self, 'layer'):
+            if isinstance(input1, Labels):
+                projection.self.viewer.add_labels(output, translate=input1.translate[1:3])
+            else:
+                projection.self.viewer.add_image(output, translate=input1.translate[1:3])
+        else:
+            projection.self.layer.data = output
+            projection.self.layer.contrast_limits = input1.contrast_limits
+            projection.self.layer.name = "Result of " + operation.__name__
+            projection.self.layer.translate = input1.translate[1:3]
